/* friedman.c
 * written by ff123
 * http://ff123.net/
 *
 * Version History
 * 1.00  23 Sept 2001   Original
 * 1.10  29 Sept 2001   Added option to perform blocked ANOVA analysis
 *                      bugfix:  reports correct confidence if -s is used
 * 1.20   8 Oct  2001   Put ranking data in separate file ("rank.txt")
 *                      Added more text in this section explaining the
 *                      analysis.  Output to screen and log file has
 *                      a different format, including an ANOVA table
 *                      if that type of analysis is run.  Added option
 *                      to perform Tukey's HSD, either parametric or
 *                      non-parametric.
 * 1.21  16 Oct  2001   Pipe error output to stdout instead of stderr
 * 1.22   1 Nov  2001   Changed DELTA to 1E-7, which is just slightly
 *                      greater than the accuracy of float.
 *                      Changed one instance of floating point comparison
 *                      in which DELTA was incorrect sign.
 * 1.23   9 Nov  2001   Changed Friedman and ANOVA matrixes to display
 *                      p-values instead of differences.  The Tukey
 *                      matrixes must remain differences.
 * 1.24  17 Jan  2002   Oops.  The p-value for the Friedman was being
 *                      calculated incorrectly.  Fixed the formula
 *                      within fried_pval() by multiplying by 2.0.
 *
 *
 * Default Friedman Analysis:
 *   Performs a nonparametric analysis on data.  The data is converted
 *   to the form of ranks (as opposed to ratings), where each row of
 *   the table contains the ranks assigned to the samples by each judge.
 *   The columns are the samples being judged.  The analysis is performed
 *   using a Friedman-type statistic.  The results are reasonably
 *   accurate for studies involving 12 or more listeners.  This type
 *   of analysis assumes the least about the data:  that is, normal
 *   distribution of listeners and equal interval rating scale are *not*
 *   assumed.
 *
 *   Multiple Comparison Procedure
 *     If the Friedman analysis determines that not all the samples
 *     were equally rated, then a (protected) non-parametric Fisher Least
 *     Significant Difference (LSD) statistic is calculated to perform a
 *     multiple comparison.  This is a one-at-a=time multiple comparison
 *     procedure.  That is, the Type I error (of size alpha) applies each
 *     time a comparison is made. 
 *
 *     Tukey's Honestly Significant Difference (HSD) can be run instead
 *     of the Friedman/Fisher LSD.  Tukey's HSD is a simultaneous
 *     multiple comparison procedure which aims to control the experiment-
 *     wise error and is very conservative in the sense that false
 *     differences are controlled at sometimes at the expense of missing
 *     true differences.
 *
 *
 * Optional Blocked ANOVA Analysis:
 *   Performs a blocked (correlated) ANOVA analysis, where each listener
 *   is a block.  Unlike a regular ANOVA, assumes that panelists are
 *   consistent in rating the samples, but that they might use different
 *   parts of the scale to express their perceptions.  Thus by taking
 *   into account this type of judge-to-judge difference, it yields a
 *   more accurate estimate of experimental error and thus more sensitive
 *   tests of hypotheses than the regular ANOVA.  ANOVA analyses assume
 *   more about the data than the non-parametric types, such as the
 *   Friedman.  But ANOVA is generally more sensitive/powerful in
 *   exchange.  Two important assumptions:  listeners are normally
 *   distributed and the rating scale is equal interval.
 *
 *   Multiple Comparison Procedure
 *     If the blocked ANOVA analysis determines that not all the samples
 *     were equally rated, then a (protected) Fisher Least Significant
 *     Difference (LSD) statistic is calculated to perform a multiple
 *     comparison.  This is a one-at-a=time multiple comparison procedure.
 *     That is, the Type I error (of size alpha) applies each time a
 *     comparison is made.
 *
 *     Tukey's Honestly Significant Difference (HSD) can be run instead
 *     of the ANOVA/Fisher LSD.  Tukey's HSD is a simultaneous
 *     multiple comparison procedure which aims to control the experiment-
 *     wise error and is very conservative in the sense that false
 *     differences are controlled at sometimes at the expense of missing
 *     true differences.
 *
 *
 * Uses chisq.c and z.c modules written by Gary Perlman of Wang
 * Institute, Tyngsboro, MA, to calculate chi-square and z-test
 * statistics.
 *
 * Uses fisher.c module written by Matthew Belmonte and modified by me
 * to perform F-tests and Student's t tests.
 *
 *---------------------------------------------------------------------
 *
 * The input file must have the following file format:
 *
 * % Example input file
 * % Lines that start with a percent sign are designated as comments.
 * % Also, blank lines are ignored as comments.
 *
 * SampleA  SampleB  SampleC  SampleD
 *   2.4      4.0      3.2      3.2
 *   1.3      3.4      5.0      4.0
 *   4.0      4.0      2.0      2.0
 *
 * The first non-comment line contains the labels for the samples.
 * The labels must not contain any spaces in them.
 *
 *---------------------------------------------------------------------
 * A log file is generated containing everything printed to screen
 * ("log.txt").
 *
 * Also, a file containing the data in ranked format is created
 * ("rank.txt").
 *
 * % Example output
 *
 * t1  t2  t3  t4  person  s1   s2   s3   s4
 * 1   2   0   1   1       1.0  4.0  2.5  2.5
 * 1   1   1   1   2       1.0  2.0  4.0  3.0
 * 2   0   2   0   3       3.5  3.5  1.5  1.5
 *
 * The average of the tied ranks are assigned to each of the samples
 * that could not be differentiated.  For example, the first listener
 * above assigned a tie between SampleC and SampleD.  The average of
 * the these two (normally 2 and 3) are assigned to both.
 *
 * The columns to the left of the samples columns show how many tied
 * groups there were for a particular listener.  Non-tied samples are
 * each counted as a separate group of tied size of 1.  If you wish
 * to perform the Friedman test manually from within Excel to get an
 * idea of how it works, you will find this file useful.
 *
 **********************************************************************/
#define SCOPE
#include "friedman.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>


/* static variables */
static char *infilename_s = NULL;          /* input filename */
static char *logfilename_s = "log.txt";    /* log filename */
static char *rnkfilename_s = "rank.txt";   /* rank filename */
static FILE *ifp_s = NULL;                 /* pointer to input file */
static FILE *log_s = NULL;                 /* pointer to log file */
static FILE *rnk_s = NULL;                 /* pointer to rank file */
static int analysis_s = FRIEDMAN;          /* analysis type */
static char label_s[MAXSAMPLES][MAXCHAR];  /* sample labels */
static int numlisteners_s = 0;             /* number of listeners */
static int numsamples_s = 0;               /* number of samples */
static float rawaverage_s = 0;             /* average of raw data */
static float listener_sumsq_s = 0;
static float sample_sumsq_s = 0;
static float devsq_s = 0;                  /* sumsq of rawdata deviations */
static float sample_avg_s[MAXSAMPLES];     /* sample averages */
static float ranksums_s[MAXSAMPLES];       /* ranksums */
static int tcubed_s = 0;                   /* used tied groups */
static float critsig_s = 0.05;             /* critical significance */
static float T_s;                          /* Friedman's T */
static float SigDiff_s;                    /* Fisher's LSD or Tukey's HSD */
static struct {
    float df_tot;            /* total degrees of freedom */
    float df_bl;             /* blocks (listeners) degrees of freedom */
    float df_tr;             /* treatments (samples) degrees of freedom */
    float df_err;            /* error degrees of freedom */
    float sumsq_tot;         /* sumsq of deviation from rawdata */
    float sumsq_bl;
    float sumsq_tr;
    float sumsq_err;
    float meansq_tr;
    float meansq_err;
    float F;                 /* F statistic */
    float p;                 /* probability that null hypothesis is tenable */
} anova_table_s;

/**********************************************************************/
/* init global and static variables
 */

void init(void) {
    int i;       /* index */

    log_s = fopen(logfilename_s, "w");
    if (log_s == NULL) {
        fprintf(stdout, "Can't open log file %s\n", logfilename_s);
        exit(1);
    }
    rnk_s = fopen(rnkfilename_s, "w");
    if (log_s == NULL) {
        fprintf(stdout, "Can't open rank file %s\n", logfilename_s);
        exit(1);
    }
    version(stdout);
    version(log_s);

    for (i=0; i<MAXSAMPLES; i++)
        ranksums_s[i] = 0;
        sample_avg_s[i] = 0;
    return;

    anova_table_s.df_tot = 0;
    anova_table_s.df_bl = 0;
    anova_table_s.df_tr = 0;
    anova_table_s.df_err = 0;
    anova_table_s.sumsq_tot = 0;
    anova_table_s.sumsq_bl = 0;
    anova_table_s.sumsq_tr = 0;
    anova_table_s.sumsq_err = 0;
    anova_table_s.meansq_tr = 0;
    anova_table_s.meansq_err = 0;
    anova_table_s.F = 0;
    anova_table_s.p = 0;
}

/**********************************************************************/
/* Take data from rawdata and returns the ranking data.  Also returns
 * t, the array of tied groups.
 */

int rank(float rawdata[], float rankdata[], int tdata[]) {
    int i;      /* index */
    int j;      /* index */
    int k;      /* index */
    float sum;  /* sum of rankvalues */
    struct {
        int position;
        float rawvalue;
        float rankvalue;
        int t;
    } sorted[MAXSAMPLES], temp;

    /* initialize */
    for (i=0; i<numsamples_s; i++) {
        sorted[i].position = i;
        sorted[i].rawvalue = rawdata[i];
        sorted[i].rankvalue = 0;
        sorted[i].t = 1;
    }

    /* first sort from min to max */
    for (i=0; i<numsamples_s-1; i++) {
        for (j=i+1; j<numsamples_s; j++) {
            if (sorted[j].rawvalue < sorted[i].rawvalue + DELTA) {
                temp.position = sorted[i].position;
                temp.rawvalue = sorted[i].rawvalue;
                sorted[i].position = sorted[j].position;
                sorted[i].rawvalue = sorted[j].rawvalue;
                sorted[j].position = temp.position;
                sorted[j].rawvalue = temp.rawvalue;
            }
        }
    }

    /* now assign rankvalues */
    for (i=0; i<numsamples_s; i++)
        sorted[i].rankvalue = i+1;

    i = 0;
    while (i < numsamples_s - 1) {
        sum = sorted[i].rankvalue;
        j = i+1;
        while (fabs(sorted[j].rawvalue - sorted[i].rawvalue) < DELTA) {
            sum += sorted[j].rankvalue;
            sorted[i].t++;
            sorted[j].t--;
            j++;
        }
        for (k=i; k<j; k++)
            sorted[k].rankvalue = sum / sorted[i].t;
        i = j;
    }

    /* unsort data and pack into rankdata[] */
    for (i=0; i<numsamples_s; i++) {
        rankdata[sorted[i].position] = sorted[i].rankvalue;
        tdata[i] = sorted[i].t;
    }

    return 0;
}

/**********************************************************************/
/* Reads the input file.  Format is as described at the top of this
 * file.  Lines starting with percent (%) or which are blank are
 * skipped over.  Tokens are separated by commas, spaces, tabs, line
 * returns, and newlines.  Calls rank() to put the raw data into ranked
 * format.  This function must be called even when performing an ANOVA
 * analysis.  It reads the file all the way through so that an average
 * value for all the data can be calculated.
 */

int read_input(void) {
    char line[MAXCHAR];          /* string buffer */
    char *p;                     /* generic pointer */
    char buf[MAXCHAR];           /* string buffer */
    int labels_parsed = FALSE;   /* flag for labels parsed */
    float rawdata[MAXSAMPLES];   /* raw data */
    float rankdata[MAXSAMPLES];  /* ranked data */
    int tdata[MAXSAMPLES];       /* tied groups */
    int count;                   /* count of data */
    int linenum = 0;             /* line number */
    int i;                       /* index */


    ifp_s = fopen(infilename_s, "r");
    if (ifp_s == NULL) {
        fprintf(stdout, "Can't open input file %s\n", infilename_s);
        fprintf(log_s, "Can't open input file %s\n", infilename_s);
        exit(1);
    }

    numlisteners_s = 0;
    while (fgets(line, MAXCHAR, ifp_s)) {
        p = strtok(line, ", \t\r\n");
        linenum++;

        /* ignore comments and blank lines */
        if (p) {
            strncpy(buf, p, 1);
            buf[1] = '\0';
            if (buf[0] == '%')
                continue;
        }
        else
            continue;

        /* parse labels line and determine numsamples_s */
        if (labels_parsed == FALSE) {
            strcpy(label_s[0], p);
            numsamples_s++;

            while (p = strtok(NULL, ", \t\r\n")) {
                strcpy(label_s[numsamples_s], p);
                numsamples_s++;
            }
            if (numsamples_s > MAXSAMPLES) {
                fprintf(stdout, "Too many samples: %d\n", numsamples_s);
                fprintf(stdout, "Max samples allowed: %d\n", MAXSAMPLES);
                fprintf(log_s, "Too many samples: %d\n", numsamples_s);
                fprintf(log_s, "Max samples allowed: %d\n", MAXSAMPLES);
                exit(1);
            }

            if (analysis_s == FRIEDMAN) {
                /* write to rank file */
                fprintf(rnk_s, "\n");
                for (i=0; i<numsamples_s; i++)
                    fprintf(rnk_s, "t%d  ", i+1);
                fprintf(rnk_s, "person  ");
                for (i=0; i<numsamples_s; i++)
                    fprintf(rnk_s, "s%d   ", i+1);
                fprintf(rnk_s, "\n");
            }

            labels_parsed = TRUE;
            continue;
        }

        /* parse data */
        count = 0;
        strcpy(buf, p);
        rawdata[0] = atof(buf);
        count++;

        while (p = strtok(NULL, ", \t\r\n")) {
            strcpy(buf, p);
            rawdata[count] = atof(buf);
            count++;
        }
        if (count != numsamples_s) {
            fprintf(stdout, "Data in line number %d discarded; "
                    "%d samples counted in this row; should be %d\n",
                    linenum, count, numsamples_s);
            fprintf(log_s, "Data in line number %d discarded; "
                    "%d samples counted in this row; should be %d\n",
                    linenum, count, numsamples_s);
            continue;
        }
        numlisteners_s++;

        if ((analysis_s == FRIEDMAN) || (analysis_s == TUKEYNP)) {

            /* rank data */
            rank(rawdata, rankdata, tdata);

            /* write to rank file */
            for (i=0; i<numsamples_s; i++)
                fprintf(rnk_s, "%-4d", tdata[i]);
            fprintf(rnk_s, "%-8d", numlisteners_s);
            for (i=0; i<numsamples_s; i++)
                fprintf(rnk_s, "%-5.1f", rankdata[i]);
            fprintf(rnk_s, "\n");
        }

        /* compute rank sums and tcubed_s; also sum rawdata so that
         * the averages can be computed for ANOVA analysis, if performed
         */
        for (i=0; i<numsamples_s; i++) {
            ranksums_s[i] += rankdata[i];
            tcubed_s += tdata[i] * tdata[i] * tdata[i];
            sample_avg_s[i] += rawdata[i];
        }
        tcubed_s -= numsamples_s;
    }

    for (i=0; i<numsamples_s; i++) {
        sample_avg_s[i] /= numlisteners_s;
        rawaverage_s += sample_avg_s[i];
    }

    rawaverage_s /= numsamples_s;
    return 0;
}

/**********************************************************************/
/* process command line options
 * Returns INCR if switch option is followed by another argument;
 * Otherwise returns NO_INCR
 */

int process_options(int argnum, char *argv[]) {
    char option[MAXCHAR];       /* option */
    char optionvalue[MAXCHAR];  /* value of option */

    sprintf(option, "%s", *(argv + argnum));

    if (strcmp(option, "-h") == 0)
        usage_error();
    else {
        if (strcmp(option, "-s") == 0) {
            sprintf(optionvalue, "%s", *(argv + argnum + 1));
            critsig_s = atof(optionvalue);
            return INCR;
        }
        if (strcmp(option, "-a") == 0) {
            analysis_s = ANOVA;
            return NO_INCR;
        }
        if (strcmp(option, "-tp") == 0) {
            analysis_s = TUKEYP;
            return NO_INCR;
        }
        if (strcmp(option, "-tn") == 0) {
            analysis_s = TUKEYNP;
            return NO_INCR;
        }
    }

    return NO_INCR;
}

/**********************************************************************/

void usage_error(void) {
    fprintf(stdout,
        "Usage: friedman [options] <infile>\n");
    fprintf(stdout,
        "  -h               usage\n");
    fprintf(stdout,
        "  -a               Replace Friedman with a blocked ANOVA analysis\n");
    fprintf(stdout,
        "  -tp              Replace Friedman/Fisher with a parametric Tukey's HSD.\n");
    fprintf(stdout,
        "  -tn              Replace Friedman/Fisher with a non-parametric Tukey's HSD.\n");
    fprintf(stdout,
        "  -s <crit sig>    Critical significance of analysis.  Default is 0.05.\n"
        "                     Permitted values are 0.05, 0.1, and 0.01\n");
    exit(1);
}

/**********************************************************************/
/* user must enter input filename
 */

int parse_command_line(int argc, char *argv[]) {
    int i;
    int increment = FALSE;   /* increment argument counter */

    if (argc < 2) {
        fprintf(stdout, "Not enough arguments specified\n");
        usage_error();
    }

    for (i=1; i<argc; i++) {
        if (argv[i][0] == '-') {
            increment = process_options(i, argv);
            if (increment) {
                i++;
                continue;
            }
        }
        else
            if (infilename_s == NULL)
                infilename_s = argv[i];
            else {
                fprintf(stdout, "Too many arguments specified\n");
                usage_error();
            }
    }
    if (infilename_s == NULL) {
        fprintf(stdout, "Input filename not specified\n");
        usage_error();
    }

    return 0;
}

/**********************************************************************/
/* Calculate Friedman's T statistic, allowing for tied groups.
 * Reference Hollander and Wolfe, 1973.
 */

float calc_T(void) {
    float Tfinal;           /* calculated T */
    float G;                /* intermediate result */
    float numerator = 0;    /* numerator of T */
    float denominator = 0;  /* denominator of T */
    int i;                  /* index */

    G = numsamples_s * numlisteners_s * (numsamples_s + 1) / 2.0;
    for (i=0; i<numsamples_s; i++) {
        numerator += pow(ranksums_s[i] - G/numsamples_s, 2.0);
    }
    numerator *= 12.0;

    denominator = numlisteners_s * numsamples_s * (numsamples_s + 1) -
                  (1.0 / (numsamples_s - 1)) * tcubed_s;

    Tfinal = numerator / denominator;
    return Tfinal;
}

/**********************************************************************/
/* Calculate F statistic for blocked ANOVA analysis.  Assumes input
 * file is still open, that is, read_input() has already been called.
 */

void calc_F(void) {
    char line[MAXCHAR];          /* string buffer */
    char *p;                     /* generic pointer */
    char buf[MAXCHAR];           /* string buffer */
    int labels_parsed = FALSE;   /* flag for labels parsed */
    float rawdata[MAXSAMPLES];   /* raw data */
    float listener_avg = 0;      /* listener averages */
    int count;                   /* count of data */
    int linenum = 0;             /* line number */
    int i;                       /* index */


    rewind(ifp_s);

    numlisteners_s = 0;
    while (fgets(line, MAXCHAR, ifp_s)) {
        p = strtok(line, ", \t\r\n");
        linenum++;

        /* ignore comments and blank lines */
        if (p) {
            strncpy(buf, p, 1);
            buf[1] = '\0';
            if (buf[0] == '%')
                continue;
        }
        else
            continue;

        /* read labels line, then continue */
        if (labels_parsed == FALSE) {
            labels_parsed = TRUE;
            continue;
        }

        /* parse data */
        count = 0;
        strcpy(buf, p);
        rawdata[0] = atof(buf);
        count++;

        while (p = strtok(NULL, ", \t\r\n")) {
            strcpy(buf, p);
            rawdata[count] = atof(buf);
            count++;
        }
        if (count != numsamples_s) {
            fprintf(stdout, "Data discarded for listener %d; "
                    "%d samples counted in this row; should be %d\n",
                    linenum, count, numsamples_s);
            fprintf(log_s, "Data discarded for listener %d; "
                    "%d samples counted in this row; should be %d\n",
                    linenum, count, numsamples_s);
            continue;
        }
        numlisteners_s++;

        /* calculate listener averages */
        for (i=0, listener_avg=0; i<numsamples_s; i++) {
            listener_avg += rawdata[i];
            anova_table_s.sumsq_tot += pow((rawdata[i] - rawaverage_s), 2);
        }
        listener_avg /= numsamples_s;
        listener_sumsq_s += pow((listener_avg - rawaverage_s), 2);        
    }
    for (i=0, sample_sumsq_s=0; i<numsamples_s; i++) {
        sample_sumsq_s += pow((sample_avg_s[i] - rawaverage_s), 2);
    }
    anova_table_s.df_tot = numsamples_s * numlisteners_s - 1;
    anova_table_s.df_bl = numlisteners_s -1;
    anova_table_s.df_tr = numsamples_s - 1;
    anova_table_s.df_err = anova_table_s.df_bl * anova_table_s.df_tr;
    anova_table_s.sumsq_bl = numsamples_s * listener_sumsq_s;
    anova_table_s.sumsq_tr = numlisteners_s * sample_sumsq_s;
    anova_table_s.sumsq_err = anova_table_s.sumsq_tot -
                    anova_table_s.sumsq_bl - anova_table_s.sumsq_tr;
    anova_table_s.meansq_tr = anova_table_s.sumsq_tr / anova_table_s.df_tr;
    anova_table_s.meansq_err = anova_table_s.sumsq_err / anova_table_s.df_err;
    anova_table_s.F = anova_table_s.meansq_tr / anova_table_s.meansq_err;
    anova_table_s.p = 1 - fisher(anova_table_s.df_tr,
                             anova_table_s.df_err, anova_table_s.F);
    return;
}

/**********************************************************************/
/* Separate the means with an LSD or HSD procedure
 */

float calcSD(void) {
    float SDfinal = 0;         /* calculated SD */

    switch (analysis_s) {
        case (FRIEDMAN):
            SDfinal = critz(1 - critsig_s/2) *
                               sqrt(numlisteners_s * numsamples_s *
                               (numsamples_s + 1) / 6.0);
            break;
        case (ANOVA):
            SDfinal = critt(anova_table_s.df_err, 1 - critsig_s) *
                    sqrt(2 * anova_table_s.meansq_err / numlisteners_s);
            break;
        case (TUKEYP):
            SDfinal = get_tukey_q(critsig_s, numsamples_s,
                         anova_table_s.df_err) *
                         sqrt(anova_table_s.meansq_err / numlisteners_s);
            break;
        case (TUKEYNP):
            SDfinal = get_tukey_q(critsig_s, numsamples_s, INFINITY) *
                         sqrt(numlisteners_s * numsamples_s *
                         (numsamples_s + 1) / 12.0);
            break;
        default:
            break;
    }

    return SDfinal;
}

/**********************************************************************/
/* calculate p-values for Fisher's LSD, given the differences in ranksums
 */

float fried_pval(float ranksumdiff) {
    float pval;

    pval = 2.0 * (1 - poz(ranksumdiff/sqrt(numlisteners_s *
                      numsamples_s * (numsamples_s + 1) / 6.0)));
    return pval;
}

/**********************************************************************/
/* calculate p-values for Fisher's LSD, given the differences in means
 */

float anova_pval(float meandiff) {
    float pval;

    pval = 1 - student(anova_table_s.df_err,
           meandiff/sqrt(2 * anova_table_s.meansq_err / numlisteners_s));
    return pval;
}

/**********************************************************************/
/* construct Fisher LSD or Tukey's HSD matrix.  First sort data from
 * highest to lowest, then group them together.
 */

void construct_sd_matrix(FILE *output) {
    samples_t sorted;         /* data to be sorted */
    samples_t *sortedp;       /* pointer to sorted data */
    int i;                    /* index */
    int j;                    /* index */
    char buf[MAXCHAR];        /* string buffer */
    char buf2[MAXCHAR];       /* string buffer */
    float lsdmatrix[MAXSAMPLES][MAXSAMPLES];
    int is_better;            /* flag */

    /* init data structures */
    for (i=0; i<numsamples_s; i++) {
        if ((analysis_s == FRIEDMAN) || (analysis_s == TUKEYNP))
            sorted.data[i] = ranksums_s[i];
        else
            sorted.data[i] = sample_avg_s[i];
        sorted.num[i] = i;
    }
    sortedp = &sorted;
    sort(sortedp, numsamples_s);

    /* now print out all scores */
    switch (analysis_s) {
        case FRIEDMAN:  /* flow thru */
        case TUKEYNP:
            fprintf(output, "Ranksums:\n");
            break;
        case ANOVA:     /* flow thru */
        case TUKEYP:
            fprintf(output, "Means:\n");
            break;
        default:
            break;
    }
    fprintf(output, "\n");
    for (i=0; i<numsamples_s; i++) {
        sprintf(buf, label_s[sorted.num[i]]);
        buf[8] = '\0';
        fprintf(output, "%-9s", buf);
    }
    fprintf(output, "\n");
    for (i=0; i<numsamples_s; i++)
        fprintf(output, "%6.2f   ", sorted.data[i]);
    fprintf(output, "\n\n");

    for (i=0; i<numsamples_s-1; i++)
        for (j=i+1; j<numsamples_s; j++)
            lsdmatrix[i][j] = sorted.data[i] - sorted.data[j];
    if ((analysis_s == FRIEDMAN) || (analysis_s == ANOVA))
        fprintf(output,
        "---------------------------- p-value Matrix ---------------------------\n\n");
    else
        fprintf(output,
        "-------------------------- Difference Matrix --------------------------\n\n");

    fprintf(output, "%-9s", "");
    for (i=1; i<numsamples_s; i++) {
        sprintf(buf, label_s[sorted.num[i]]);
        buf[8] = '\0';
        fprintf(output, "%-9s", buf);
    }
    fprintf(output, "\n");

    for (i=0; i<numsamples_s-1; i++) {
        sprintf(buf, label_s[sorted.num[i]]);
        buf[8] = '\0';
        fprintf(output, "%-9s", buf);
        for (j=1; j<numsamples_s; j++) {
            if (j <= i)
                fprintf(output, "%-9s", "");
            else {
                if (lsdmatrix[i][j] >= SigDiff_s - DELTA) {
                    switch (analysis_s) {
                        case FRIEDMAN:
                            fprintf(output, "%-5.3f*   ",
                                    fried_pval(lsdmatrix[i][j]));
                            break;
                        case ANOVA:
                            fprintf(output, "%-5.3f*   ",
                                    anova_pval(lsdmatrix[i][j]));
                            break;
                        case TUKEYNP:  /* flow-thru */
                        case TUKEYP:
                            fprintf(output, "%7.3f* ", lsdmatrix[i][j]);
                            break;
                        default:
                            break;
                    }
                }
                else {
                    switch (analysis_s) {
                        case FRIEDMAN:
                            fprintf(output, "%-5.3f    ",
                                    fried_pval(lsdmatrix[i][j]));
                            break;
                        case ANOVA:
                            fprintf(output, "%-5.3f    ",
                                    anova_pval(lsdmatrix[i][j]));
                            break;
                        case TUKEYNP:  /* flow-thru */
                        case TUKEYP:
                            fprintf(output, "%7.3f  ", lsdmatrix[i][j]);
                            break;
                        default:
                            break;
                    }
                }
            }
        }
        fprintf(output, "\n");
    }
    fprintf(output, "-----------------------------------------------------------------------\n\n");

    for (i=0; i<numsamples_s-1; i++) {
        is_better = FALSE;
        sprintf(buf, "%s is better than ", label_s[sorted.num[i]]);
        for (j=i+1; j<numsamples_s; j++) {
            if (lsdmatrix[i][j] >= SigDiff_s - DELTA) {
                sprintf(buf2, "%s, ", label_s[sorted.num[j]]);
                strcat(buf, buf2);
                is_better = TRUE;
            }
        }
        /* remove comma from end of line */
        buf[strlen(buf)-2] = '\0';
        if (is_better)
            fprintf(output, "%s\n", buf);
    }
}

/**********************************************************************/
/* assesses significance of Friedman's T or ANOVA F.  Returns TRUE if
 * significant, FALSE if not significant.
 */

int assess_significance(FILE *output) {
    float sig;                 /* significance of Friedman's T */
    int is_signif = FALSE;     /* return value */

    if (analysis_s == FRIEDMAN)
        sig = pochisq(T_s, numsamples_s-1);
    else
        sig = anova_table_s.p;

    fprintf(output, "Significance of data: %5.2E ", sig);
    if ((sig <= critsig_s + DELTA) && (sig <= 0.01 + DELTA)) {
        fprintf(output, "(highly significant)\n");
        is_signif = TRUE;
    }
    if ((sig <= critsig_s + DELTA) && (sig <= 0.05 + DELTA) && (sig > 0.01 + DELTA)) {
        fprintf(output, "(significant)\n");
        is_signif = TRUE;
    }
    if ((sig <= critsig_s + DELTA) && (sig > 0.05 + DELTA)) {
        fprintf(output, "\n");
        is_signif = TRUE;
    }
    if (sig > critsig_s + DELTA)
        fprintf(output, "(not significant)\n");

    return is_signif;
}

/**********************************************************************/
/* outputs info either to SCREEN or LOGFILE
 */

void print_info(int device) {
    int i;              /* index */
    int j;              /* index */
    FILE *output;       /* file pointer */
    int do_fisher;      /* TRUE if omnibus test is significant */

    if (device == SCREEN)
        output = stdout;
    if (device == LOGFILE)
        output = log_s;

    switch (analysis_s) {
        case (FRIEDMAN):
            fprintf(output, "%s\n", "Friedman Analysis");
            fprintf(output, "\n");
            fprintf(output, "Number of listeners: %d\n", numlisteners_s);
            fprintf(output, "Critical significance:  %4.2f\n", critsig_s);
            do_fisher = assess_significance(output);
            if (do_fisher == TRUE) {
                fprintf(output, "Fisher's protected LSD for %s: %7.3f\n\n",
                        (analysis_s == FRIEDMAN) ? "rank sums" : "ANOVA", SigDiff_s);
                construct_sd_matrix(output);
            }
            break;
        case (ANOVA):
            fprintf(output, "%s\n", "Blocked ANOVA analysis");
            fprintf(output, "\n");
            fprintf(output, "Number of listeners: %d\n", numlisteners_s);
            fprintf(output, "Critical significance:  %4.2f\n", critsig_s);
            do_fisher = assess_significance(output);
            fprintf(output,
                     "---------------------------------------------------------------\n");
            fprintf(output, "ANOVA Table for Randomized Block Designs Using Ratings\n\n");
            fprintf(output, "Source of         Degrees     Sum of    Mean\n");
            fprintf(output,
                     "variation         of Freedom  squares   Square    F      p\n");
            fprintf(output, "\n");
            fprintf(output, "Total             %4.0f        %7.2f\n",
                     anova_table_s.df_tot, anova_table_s.sumsq_tot);
            fprintf(output, "Testers (blocks)  %4.0f        %7.2f\n",
                     anova_table_s.df_bl, anova_table_s.sumsq_bl);
            fprintf(output, "Codecs eval'd     %4.0f        %7.2f  %6.2f   %5.2f  %8.2E\n",
                     anova_table_s.df_tr, anova_table_s.sumsq_tr,
                     anova_table_s.meansq_tr, anova_table_s.F, anova_table_s.p);
            fprintf(output, "Error             %4.0f        %7.2f  %6.2f\n",
                     anova_table_s.df_err, anova_table_s.sumsq_err,
                     anova_table_s.meansq_err);
            fprintf(output,
                     "---------------------------------------------------------------\n");
            if (do_fisher == TRUE) {
                fprintf(output, "Fisher's protected LSD for %s: %7.3f\n\n",
                        (analysis_s == FRIEDMAN) ? "rank sums" : "ANOVA", SigDiff_s);
                construct_sd_matrix(output);
            }
            break;
        case (TUKEYP):
            fprintf(output, "%s\n", "Tukey HSD analysis");
            fprintf(output, "\n");
            fprintf(output, "Number of listeners: %d\n", numlisteners_s);
            fprintf(output, "Critical significance:  %4.2f\n", critsig_s);
            fprintf(output, "Tukey's HSD: %7.3f\n\n", SigDiff_s);
            construct_sd_matrix(output);
            break;
        case (TUKEYNP):
            fprintf(output, "%s\n", "Nonparametric Tukey HSD analysis");
            fprintf(output, "\n");
            fprintf(output, "Number of listeners: %d\n", numlisteners_s);
            fprintf(output, "Critical significance:  %4.2f\n", critsig_s);
            fprintf(output, "Nonparametric Tukey's HSD: %7.3f\n\n", SigDiff_s);
            construct_sd_matrix(output);
            break;
        default:
            break;
    }
}

/**********************************************************************/
/* prints version to screen or to file
 */

void version(FILE *output) {
    fprintf(output, "FRIEDMAN version %s\n", VERSION);
}

/**********************************************************************/

int main(int argc, char *argv[]) {
    init();
    parse_command_line(argc, argv);
    read_input();
    switch (analysis_s) {
        case (FRIEDMAN):
            /* flow-through */
        case (TUKEYNP):
            T_s = calc_T();
            break;
        case (ANOVA):
            /* flow-through */
        case (TUKEYP):
            calc_F();
            break;
        default:
            break;
    }
    SigDiff_s = calcSD();

    print_info(SCREEN);
    print_info(LOGFILE);

    if (rnk_s == NULL) fclose(rnk_s);
    if (log_s == NULL) fclose(log_s);
    if (ifp_s == NULL) fclose(ifp_s);
    return 0;
}
